/*
 * mipmap_fct.h
 *
 * My own private optimized gluBuild2DMipmaps() function.
 *
 * This piece of code builds a mipmap and sends it to OpenGL.
 * It is intended to be fast: 2 to 4 times faster than the default
 * function that uses gluScaleImage(). It eats less memory too.
 * mipmap.c includes it.
 *
 * Stephane Rehel, October 24 1998
 */

/*
 * parameters:
 *    GLenum target
 *    GLint width, GLint height
 *    GLvoid* data
 *    GLboolean in_place
 *
 * width and height are both a power of 2 and > 0
 * in_place==TRUE if scaling can be done in place
 *
 * Defines:
 *   _GL_FORMAT : format (e.g. GL_RGBA)
 *   COMPS      : number of components (e.g. 4 for GL_RGBA)
 *   _GL_TYPE   : type of single components (e.g. GL_UNSIGNED_BYTE)
 *   TYPE       : type of single components (e.g. unsigned char)
 *   TTYPE      : intermediate type for scaling (e.g int )
 */

{
  TYPE* image;
  int level= 0;

  if( ! in_place )
    {
    /* we cannot scale the image in place */
    image= (TYPE*) malloc( width * height * COMPS * sizeof(TYPE) );
    MEMCPY( image, data, width * height * COMPS * sizeof(TYPE) );
    }
   else
    image= (TYPE*) data;

  /* level 0 */
  glTexImage2D( target, level++, 3 /*GL_RGB*/, width, height, 0,
                _GL_FORMAT, _GL_TYPE, (GLvoid*) image );

  /* mmw at least should be a constant */
  #define xy_scale(mmw,mmh)                                            \
      {                                                                \
      register int j;                                                  \
      register TYPE* q= image;                                         \
      register TYPE* r= image;                                         \
      for( j= mmh; j; --j, r += mmw*2*COMPS * 2, q += mmw*COMPS )      \
        {                                                              \
        register int i;                                                \
        for( i= mmw; i; --i )                                          \
          {                                                            \
          q[i*COMPS-COMPS+0]= (TYPE)(                                  \
            ( (TTYPE)r[i*COMPS*2-COMPS*2+0*COMPS+0]                    \
            + (TTYPE)r[i*COMPS*2-COMPS*2+1*COMPS+0]                    \
            + (TTYPE)r[i*COMPS*2-COMPS*2+0*COMPS+0+mmw*2*COMPS]        \
            + (TTYPE)r[i*COMPS*2-COMPS*2+1*COMPS+0+mmw*2*COMPS]) / 4 );\
          q[i*COMPS-COMPS+1]= (TYPE)(                                  \
            ( (TTYPE)r[i*COMPS*2-COMPS*2+0*COMPS+1]                    \
            + (TTYPE)r[i*COMPS*2-COMPS*2+1*COMPS+1]                    \
            + (TTYPE)r[i*COMPS*2-COMPS*2+0*COMPS+1+mmw*2*COMPS]        \
            + (TTYPE)r[i*COMPS*2-COMPS*2+1*COMPS+1+mmw*2*COMPS]) / 4 );\
          q[i*COMPS-COMPS+2]= (TYPE)(                                  \
            ( (TTYPE)r[i*COMPS*2-COMPS*2+0*COMPS+2]                    \
            + (TTYPE)r[i*COMPS*2-COMPS*2+1*COMPS+2]                    \
            + (TTYPE)r[i*COMPS*2-COMPS*2+0*COMPS+2+mmw*2*COMPS]        \
            + (TTYPE)r[i*COMPS*2-COMPS*2+1*COMPS+2+mmw*2*COMPS]) / 4 );\
          }                                                            \
        }                                                              \
      }

  /* mmh == 1 */
  #define x_scale(mmw)                                                 \
      {                                                                \
      register TYPE* q= image;                                         \
      register int i;                                                  \
      for( i= mmw; i; --i )                                            \
        {                                                              \
        q[i*COMPS-COMPS+0]= (TYPE)(                                    \
          ( (TTYPE)q[i*COMPS*2-COMPS*2+0*COMPS+0]                      \
          + (TTYPE)q[i*COMPS*2-COMPS*2+1*COMPS+0] ) / 2);              \
        q[i*COMPS-COMPS+1]= (TYPE)(                                    \
          ( (TTYPE)q[i*COMPS*2-COMPS*2+0*COMPS+1]                      \
          + (TTYPE)q[i*COMPS*2-COMPS*2+1*COMPS+1] ) / 2);              \
        q[i*COMPS-COMPS+2]= (TYPE)(                                    \
          ( (TTYPE)q[i*COMPS*2-COMPS*2+0*COMPS+2]                      \
          + (TTYPE)q[i*COMPS*2-COMPS*2+1*COMPS+2] ) / 2);              \
        }                                                              \
      }

  /* mmw == 1 */
  #define y_scale(mmh)                                                 \
      {                                                                \
      register int j;                                                  \
      register TYPE* q= image;                                         \
      for( j= mmh; j; --j )                                            \
        {                                                              \
        q[j*COMPS-COMPS+0]= (TYPE)(                                    \
          ( (TTYPE)q[j*COMPS*2-COMPS*2+0*COMPS+0]                      \
          + (TTYPE)q[j*COMPS*2-COMPS*2+0*COMPS+0+ 1 *2*COMPS] ) / 2 ); \
        q[j*COMPS-COMPS+1]= (TYPE)(                                    \
          ( (TTYPE)q[j*COMPS*2-COMPS*2+0*COMPS+1]                      \
          + (TTYPE)q[j*COMPS*2-COMPS*2+0*COMPS+1+ 1 *2*COMPS] ) / 2 ); \
        q[j*COMPS-COMPS+2]= (TYPE)(                                    \
          ( (TTYPE)q[j*COMPS*2-COMPS*2+0*COMPS+2]                      \
          + (TTYPE)q[j*COMPS*2-COMPS*2+0*COMPS+2+ 1 *2*COMPS] ) / 2 ); \
        }                                                              \
      }

  #define set_level(mmw,mmh)                                           \
      {                                                                \
      glTexImage2D( target, level++, 3 /*GL_RGB*/,                     \
                    mmw, mmh, 0,                                       \
                    _GL_FORMAT, _GL_TYPE, (GLvoid*) image );           \
      }

  if( width == height )
    {
    while( width > 1024 )
      {
      width /= 2;
      xy_scale(width,width)
      set_level(width,width)
      }

    /* #define it so that mmw could be a constant in *_scale() */
    #define set(s)                                                \
      if( width == s )                                            \
        {                                                         \
        width= s/2;                                               \
        xy_scale(s/2,s/2)                                         \
        set_level(s/2,s/2)                                        \
        }
    set(1024) set(512) set(256) set(128)
    set(  64) set( 32) set( 16) set(  8)
    set(   4) set(  2)
    #undef set
    }
   else
  if( width > height )
    {
    while( width > 1024 )
      {
      width /= 2;
      if( height == 1 )
        { x_scale(width) }
       else
        { height /= 2; xy_scale(width,height)  }
      set_level(width,height)
      }

    /* #define it so that mmw could be a constant in *_scale() */
    #define set(s)                                                \
       if( width == s )                                           \
         {                                                        \
         width= s/2;                                              \
         if( height == 1 ) { x_scale(s/2) }                       \
                      else { height /= 2; xy_scale(s/2,height) }  \
         set_level(s/2,height)                                    \
         }
    set(1024) set(512) set(256) set(128)
    set(  64) set( 32) set( 16) set(  8)
    set(   4) set(  2)
    #undef set
    }
   else
    {
    /* width < height */
    while( width > 1024 )
      {
      width /= 2; height /= 2;
      xy_scale(width,height)
      set_level(width,height)
      }

    /* #define it so that mmw could be a constant in *_scale() */
    #define set(s)                                                \
       if( width == s )                                           \
         {                                                        \
         width= s/2; height /= 2;                                 \
         xy_scale(s/2,height)                                     \
         set_level(s/2,height)                                    \
         }
    set(1024) set(512) set(256) set(128)
    set(  64) set( 32) set( 16) set(  8)
    set(   4) set(  2)
    #undef set

    while( height > 1 )
      {
      height /= 2;
      y_scale(height)
      set_level(1,height)
      }
    }

#undef set_level
#undef y_scale
#undef x_scale
#undef xy_scale

  if( ! in_place )
    free(image);
}
